var variables = {};
var constraints = [];
var feData;
var feProperties = {};

const specialCharacters = ["(", ")", "/", "*", "+", "-", "%", "Math", ".", "pow"];
const MACRO_KEY_PROPERTIES = ".prop.";
const MACRO_KEY_FEDATA = "FETYPE";

function buildRules(){
  var builder = new hd.ComponentBuilder();
  builder.variables(getVariables().join(","))
  for(var i in constraints){
    var constraint = constraints[i];
    builder.constraint(constraint.constraint)
      .method(constraint.methodString, constraint.methodFunction);
  }
  printRulesToConsole();
  return builder.spec();
}

function addSetOfRules(ruleSet){
  var isRules = false;
  var rules = ruleSet.split(/\n/);
  parsePropertiesMacro(rules);
  for(var i in rules){
    if(rules[i] && rules[i].trim().length > 0 && !rules[i].includes("//")){
      var rule = splitRuleAndResult(rules[i]);
      addRule(rule[0], rule[1]);
      isRules = true;
    }
  }
  return isRules;
}

function addRule(result, rule){
  if(!rule.includes("function")){
    parseVariables(rule);
  }
  addVariable(result);
  addMethod(result, rule);
}

function addMethod(result, inRule){
  if(inRule.includes("function")) {
    addFunctionMethod(result, inRule);
  } else if(inRule.includes(MACRO_KEY_FEDATA)){
    addMacroMethod(result, inRule);
  } else {
    addNonMacroMethod(result, inRule);
  }
}

function addMacroMethod(result, inRule){
  var feType = getFeType(inRule);
  var rule = "";
  for(var i in feData) {
    if (inRule.includes(MACRO_KEY_PROPERTIES) && feData[i].model.includes(feType)){
      if(feProperties[feType]){
        for(var j in feProperties[feType]){
          var subRule = feData[i].id + "PROP" + feProperties[feType][j];
          if(inRule.includes(feProperties[feType][j])){
            rule += subRule + "+";
          }
          addVariable(removeNumbers(subRule));
        }
      }
    } else if(feData[i].model.split(".").join("").includes(feType)){
      rule += feData[i].id + "+";
      if(feProperties[feType]){
        for(var j in feProperties[feType]){
          addVariable(feData[i].id + "PROP" + feProperties[feType][j]);//TODO Adds properties like uuid.prop, have to check if "." is accepted in hotdrink
        }
      }
    }
  }
  var variablesExceptResult = getVariables().filter(function(value){ return value != result && rule.includes(value); })
  if(rule.length > 0){
    rule = rule.substr(0, rule.length - 1);//fjerner siste "+"
    var methodString = variablesExceptResult .join(",") + " -> " + result;
    var methodFunction = eval("(" + variablesExceptResult.join(",") + ") => " + rule);
    addConstraint(methodFunction, methodString);
  }
}

function addFunctionMethod(result, inRule){
  //test = function (){ var i = 1; while(!(i%10==0)){ i++} return i }
  parseVariables(result);
  var start = inRule.indexOf("function");
  var end = inRule.lastIndexOf("}");
  if(end){
    if(inRule.includes(MACRO_KEY_FEDATA)){
      var variablesExceptResult = getVariables().filter(function(value){ return value != result && inRule.includes(value); })
      var feListFromMacro = getFeListFromMacro(inRule);
      for(var i in feListFromMacro){
        var fe = feListFromMacro[i];
        var macroList = getMacroProperties(inRule);
        for(var j in macroList){
          var variable = fe.id + "PROP" + macroList[j];
          addVariable(variable);
          variablesExceptResult.push(variable);

          // const methodString = variable + " -> " + result;
          const methodString = " -> " + variable;
          var functionString = fe.properties[macroList[j]];
          const methodFunction = eval("(" + variable + ") => " + functionString);
          if(newConstraint(methodString)){
            addConstraint(methodFunction, methodString);
          }
        }
        //"function (FETYPE.LOAN.INSTALLMENT.prop.amount,FETYPE.LOAN.INSTALLMENT.prop.value,FETYPE.LOAN.INSTALLMENT.prop.nominalInterestRate){ return amount / value / nominalInterestRate; }"
        // lag constraint for hver macro per øe
        // map macro til denne constrainten
        const methodString = variablesExceptResult.join(",") + " -> " + result;
        var functionString = inRule.substring(start, end+1);
        functionString = replaceParameters(macroList, functionString);
        const methodFunction = eval("(" + variablesExceptResult.join(",") + ") => " + functionString + replaceParameters(variablesExceptResult,"()"));
        addConstraint(methodFunction, methodString);
      }
    } else {

      var functionString = inRule.substring(start, end+1);
      variablesExceptResult = getVariables().filter(function(value){ return value != result && inRule.includes(value); })
      const methodString = variablesExceptResult.join(",") + " -> " + result;
      functionString = replaceParameters(variablesExceptResult, functionString);
      const methodFunction = eval("(" + variablesExceptResult.join(",") + ") => " + functionString + replaceParameters(variablesExceptResult,"()"));
      addConstraint(methodFunction, methodString);
    }
  }
}
function newConstraint(methodString){
  for(var i in constraints){
    var constraint = constraints[i];
    if(constraint && constraint.methodString === methodString){
      return false;
    }
  }
  return true;
}

function addNonMacroMethod(result, rule){
  var variablesExceptResult = getVariables().filter(function(value){ return value != result && rule.includes(value); });
  var methodString = variablesExceptResult.join(",") + " -> " + result;
  var methodFunction = eval("(" + variablesExceptResult.join(",") + ") => " + rule);
  addConstraint(methodFunction, methodString);
}

function filterNonVariables(value){
  if(isNotNumber(value) && isNotSpecialCharacter(value)){
    return value;
  } else {
    return undefined;
  }

}

function isNotSpecialCharacter(value){
  for(var i in specialCharacters){
    if(value.includes(specialCharacters[i])){
      return false;
    }
  }
  return true;
}


function addConstraint(methodFunction, methodString){
  var constraint = {};
  constraint.constraint = getVariables().filter(function(value){ return methodString.includes(value); }).join(",");
  constraint.methodFunction = methodFunction;
  constraint.methodString = methodString;
  constraints.push(constraint);
}

function parsePropertiesMacro(rules, result) {
  for(var i in rules){
    var rule = rules[i];
    if(rule.includes(MACRO_KEY_PROPERTIES)){
      var feType = getFeType(rule);
      var properties = getMacroProperties(rule);
      for (var i in properties){
        if(feProperties[feType]){
          feProperties[feType].push(properties[i]);
        } else {
          feProperties[feType] = [];
          feProperties[feType].push(properties[i]);
        }
      }
    }
  }
}

function getMacroProperties(inRule){
  //"function (FETYPE.LOAN.INSTALLMENT.prop.amount,FETYPE.LOAN.INSTALLMENT.prop.value,FETYPE.LOAN.INSTALLMENT.prop.nominalInterestRate
  // amount, value, nominalInterestRate
  var result = [];
  var string = inRule;
  var index = string.indexOf(MACRO_KEY_PROPERTIES) + MACRO_KEY_PROPERTIES.length;
  for(var i = index ; i <= string.length; i++) {
    if(!isLetter(string, i)){
      result.push(string.substring(index, i));
      if(string.charAt(i) == ')'){
        break;
      } else {

      }
      string = string.substring(i, string.length);
      index = string.indexOf(MACRO_KEY_PROPERTIES) + MACRO_KEY_PROPERTIES.length;
      i = index;
    }
  }
  return result;
}

function isLetter(string, i) {
  const char = string.charCodeAt(i);
  if (!(char > 64 && char < 91) && // upper alpha (A-Z)
  !(char > 96 && char < 123) || // lower alpha (a-z)
  !char){  // no character
    return false;
  }
  return true;
};

function getFeListFromMacro(inRule){
  var feType = getFeType(inRule);
  var result = [];
  for(var i in feData) {
    if (feData[i].model.includes(feType)) {
      result.push(feData[i]);
    }
  }
  return result;
}

function parseVariables(rule) {
  //TODO Hele denne må refaktoreres, looper gjennom alle ØE for hver regel selv om den bare legger til første gang. Henger egentlig sammen med properties parsing
  // if-sjekken her var når vi antok det kunne finnes en FETYPE. Nå er det alltid med flere.
  if (!feData) {
    feData = getData(feProperties);
  }
  if (rule.includes(MACRO_KEY_FEDATA)) {
    for (var i in feData) {
      addVariable(feData[i].id);
    }
    return; // REFACTOR!!
  }
  var array = rule.split(" ");
  array = array.filter(function (value) {
    return value && value.length > 1 && isNotNumber(value);
  });
  for (var i in array) {
    addVariable(array[i]);
  }
}


function replaceParameters(variablesExceptResult, functionString) {
  //"function (FETYPE.LOAN.INSTALLMENT.prop.amount,FETYPE.LOAN.INSTALLMENT.prop.value,FETYPE.LOAN.INSTALLMENT.prop.nominalInterestRate){ return amount / value / nominalInterestRate; }"
  // because variables are ordered in the order they are parsed we can luckily replace raw without worrying about order
  var start = functionString.indexOf("(");
  var end = functionString.indexOf(")");
  return functionString.substring(0, start+1) + variablesExceptResult.join(",") + functionString.substring(end, functionString.length);
}

function addVariable(variable) {
  if(!variables[variable]){
    variables[variable] = variable;
  }
}


function getVariables(){

  var array = $.map(variables, function(value, index) {
    return [value];
  });
  return array.filter(function(value){ return filterNonVariables(value); });
}

function getFeType(inRule){
  var index = inRule.indexOf(MACRO_KEY_FEDATA);
  var result = inRule.substring(index + "FETYPE.".length, inRule.length);
  if(result.includes(MACRO_KEY_PROPERTIES)){
    var index = result.indexOf(MACRO_KEY_PROPERTIES);
    if(index > 0){
      result = result.substring(0, index);
    }
  }
  return result.split(".").join("");
}

function getDescription(id){
  for(var i in feData){
    if(feData[i].id == id){
      return feData[i].name + "-" + feData[i].model + "-" + feData[i].id;
    }
  }
  return id;
}

function getValue(id){
  var value = 0;
  for(var i in feData){ // TODO DOBBEL(trippel fra caller) FOR LØKKE! REFAKTOR!!
    if(feData[i].id == id){
      value = feData[i].value;
      break;
    } else if(id.includes("PROP") && feData[i].properties && id.includes(feData[i].id)){
      var properties = id.substring(id.indexOf("PROP") + "PROP".length, id.length);
      value = feData[i].properties[properties];
      break;
    }
  }
  return value;
}

function removeNumbers(value){
  var result = value.split(" ");
  return result[0];
}

function readRules(){
  var el = document.getElementById("inputarea");
  if(el.value && el.value.length > 0){
    return addSetOfRules(el.value);
  }
  return false;
}

function printRulesToConsole(){
  console.debug("NOW PRINTING SET OF RULES:");
  console.debug(".variables(\"" + getVariables().join(",") + "\")");
  for(var i in constraints){
    var c = constraints[i];
    console.debug(".constraint(\"" + c.constraint + "\")");
    console.debug("  .method(\"" + c.methodString + "\",");
    console.debug("    " + c.methodFunction +")");
  }
}

function isNotNumber(n){
  return !isNumber(n);
}

function isNumber(n) {
  return !isNaN(parseFloat(n)) && isFinite(n);
}

function splitRuleAndResult(input){
  //test = function (){ var i = 1; while(!(i%10==0)){ i++} return i }
  var result = [];
  var index = input.indexOf("=");
  result[0] = input.substring(0, index).trim();
  result[1] = input.substring(index+1, input.length).trim();
  return result;
}

function resetRules() {
  variables = {};
  constraints = [];
  feData = undefined;
  feProperties = {};
}